from omar6995_browserforge.download import DownloadIfNotExists

DownloadIfNotExists(headers=True)

from .generator import Browser, HeaderGenerator

__all__ = [
    "Browser",
    "HeaderGenerator",
]



import re
from typing import Dict, Optional

from pyppeteer.browser import Browser
from pyppeteer.page import Page

from omar6995_browserforge.fingerprints import Fingerprint
from omar6995_browserforge.injectors.utils import InjectFunction, _fingerprint, only_injectable_headers


async def NewPage(
    browser: Browser,
    fingerprint: Optional[Fingerprint] = None,
    fingerprint_options: Optional[Dict] = None,
) -> Page:
    """
    Injects a Pyppeteer browser object with a Fingerprint.

    Parameters:
        browser (Browser): The browser to create the context in
        fingerprint (Optional[Fingerprint]): The fingerprint to inject. If None, one will be generated
        fingerprint_options (Optional[Dict]): Options for the Fingerprint generator if `fingerprint` is not passed
    """
    fingerprint = _fingerprint(fingerprint, fingerprint_options)
    function = InjectFunction(fingerprint)
    # create a new page
    page = await browser.newPage()

    await page.setUserAgent(fingerprint.navigator.userAgent)

    # Pyppeteer does not support firefox, so we can ignore checks
    cdp_sess = await page.target.createCDPSession()
    await cdp_sess.send(
        'Page.setDeviceMetricsOverride',
        {
            'screenHeight': fingerprint.screen.height,
            'screenWidth': fingerprint.screen.width,
            'width': fingerprint.screen.width,
            'height': fingerprint.screen.height,
            'mobile': any(
                name in fingerprint.navigator.userAgent for name in ('phone', 'android', 'mobile')
            ),
            'screenOrientation': (
                {'angle': 0, 'type': 'portraitPrimary'}
                if fingerprint.screen.height > fingerprint.screen.width
                else {'angle': 90, 'type': 'landscapePrimary'}
            ),
            'deviceScaleFactor': fingerprint.screen.devicePixelRatio,
        },
    )
    await page.setExtraHTTPHeaders(only_injectable_headers(fingerprint.headers, 'chrome'))

    # Only set to dark mode if the Chrome version >= 76
    version = re.search('.*?/(\d+)[\d\.]+?', await browser.version())
    if version and int(version[1]) >= 76:
        await page._client.send(
            'Emulation.setEmulatedMedia',
            {'features': [{'name': 'prefers-color-scheme', 'value': 'dark'}]},
        )

    # Inject function
    await page.evaluateOnNewDocument(function)
    return page



from omar6995_browserforge.download import DownloadIfNotExists

DownloadIfNotExists(fingerprints=True, headers=True)

from omar6995_browserforge.headers import Browser

from .generator import (
    Fingerprint,
    FingerprintGenerator,
    NavigatorFingerprint,
    Screen,
    ScreenFingerprint,
    VideoCard,
)

__all__ = [
    "Browser",
    "Fingerprint",
    "FingerprintGenerator",
    "NavigatorFingerprint",
    "Screen",
    "ScreenFingerprint",
    "VideoCard",
]



import random
import zipfile
from pathlib import Path
from typing import Any, Dict, Iterable, List, Optional, Sequence, Tuple, TypeVar, Union

try:
    import orjson as json
except ImportError:
    import json

T = TypeVar('T')
Map = Union[list, tuple]


class BayesianNode:
    """
    Implementation of a single node in a bayesian network allowing sampling from its conditional distribution
    """

    def __init__(self, node_definition: Dict[str, Any]):
        self.node_definition = node_definition

    def get_probabilities_given_known_values(
        self, parent_values: Dict[str, Any]
    ) -> Dict[Any, float]:
        """
        Extracts unconditional probabilities of node values given the values of the parent nodes
        """
        probabilities = self.node_definition['conditionalProbabilities']
        for parent_name in self.parent_names:
            parent_value = parent_values.get(parent_name)
            if parent_value in probabilities.get('deeper', {}):
                probabilities = probabilities['deeper'][parent_value]
            else:
                probabilities = probabilities.get('skip', {})
        return probabilities

    def sample_random_value_from_possibilities(
        self, possible_values: List[str], probabilities: Dict[str, float]
    ) -> Any:
        """
        Randomly samples from the given values using the given probabilities
        """
        # Python natively supports weighted random sampling in random.choices,
        # but this method is much faster
        anchor = random.random()
        cumulative_probability = 0.0
        for possible_value in possible_values:
            cumulative_probability += probabilities[possible_value]
            if cumulative_probability > anchor:
                return possible_value
        # Default to first item
        return possible_values[0]

    def sample(self, parent_values: Dict[str, Any]) -> Any:
        """
        Randomly samples from the conditional distribution of this node given values of parents
        """
        probabilities = self.get_probabilities_given_known_values(parent_values)
        return self.sample_random_value_from_possibilities(
            list(probabilities.keys()), probabilities
        )

    def sample_according_to_restrictions(
        self,
        parent_values: Dict[str, Any],
        value_possibilities: Iterable[str],
        banned_values: List[str],
    ) -> Optional[str]:
        """
        Randomly samples from the conditional distribution of this node given restrictions on the possible values and the values of the parents.
        """
        probabilities = self.get_probabilities_given_known_values(parent_values)
        valid_values = [
            value
            for value in value_possibilities
            if value not in banned_values and value in probabilities
        ]
        if valid_values:
            return self.sample_random_value_from_possibilities(valid_values, probabilities)
        else:
            return None  # Equivalent to `false` in TypeScript

    @property
    def name(self) -> str:
        return self.node_definition['name']

    @property
    def parent_names(self) -> List[str]:
        return self.node_definition.get('parentNames', [])

    @property
    def possible_values(self) -> List[str]:
        return self.node_definition.get('possibleValues', [])


class BayesianNetwork:
    """
    Implementation of a bayesian network capable of randomly sampling from its distribution
    """

    def __init__(self, path: Path) -> None:
        network_definition = extract_json(path)
        self.nodes_in_sampling_order = [
            BayesianNode(node_def) for node_def in network_definition['nodes']
        ]
        self.nodes_by_name = {node.name: node for node in self.nodes_in_sampling_order}

    def generate_sample(self, input_values: Optional[Dict[str, Any]] = None) -> Dict[str, Any]:
        """
        Randomly samples from the distribution represented by the bayesian network.
        """
        if input_values is None:
            input_values = {}
        sample = input_values.copy()
        for node in self.nodes_in_sampling_order:
            if node.name not in sample:
                sample[node.name] = node.sample(sample)
        return sample

    def generate_consistent_sample_when_possible(
        self, value_possibilities: Dict[str, Iterable[str]]
    ) -> Optional[Dict[str, Any]]:
        """
        Randomly samples values from the distribution represented by the bayesian network,
        making sure the sample is consistent with the provided restrictions on value possibilities.
        Returns None if no such sample can be generated.
        """
        return self.recursively_generate_consistent_sample_when_possible({}, value_possibilities, 0)

    def recursively_generate_consistent_sample_when_possible(
        self,
        sample_so_far: Dict[str, Any],
        value_possibilities: Dict[str, Iterable[str]],
        depth: int,
    ) -> Optional[Dict[str, Any]]:
        """
        Recursively generates a random sample consistent with the given restrictions on possible values.
        """
        if depth == len(self.nodes_in_sampling_order):
            return sample_so_far
        node = self.nodes_in_sampling_order[depth]
        banned_values: List[str] = []
        sample_value = None
        while True:
            sample_value = node.sample_according_to_restrictions(
                sample_so_far,
                value_possibilities.get(node.name, node.possible_values),
                banned_values,
            )
            if sample_value is None:
                break
            sample_so_far[node.name] = sample_value
            next_sample = self.recursively_generate_consistent_sample_when_possible(
                sample_so_far, value_possibilities, depth + 1
            )
            if next_sample is not None:
                return next_sample
            banned_values.append(sample_value)
            del sample_so_far[node.name]
        return None


def array_intersection(a: Sequence[T], b: Sequence[T]) -> List[T]:
    """
    Performs a set "intersection" on the given (flat) arrays
    """
    set_b = set(b)
    return [x for x in a if x in set_b]


def array_zip(a: List[Tuple[T, ...]], b: List[Tuple[T, ...]]) -> List[Tuple[T, ...]]:
    """
    Combines two arrays into a single array using the set union
    Args:
        a: First array to be combined.
        b: Second array to be combined.
    Returns:
        Zipped (multi-dimensional) array.
    """
    return [tuple(set(x).union(y)) for x, y in zip(a, b)]


def undeeper(obj: Dict[str, Any]) -> Dict[str, Any]:
    """
    Removes the "deeper/skip" structures from the conditional probability table
    """
    if not isinstance(obj, dict):
        return obj
    result: Dict[str, Any] = {}
    for key, value in obj.items():
        if key == 'skip':
            continue
        if key == 'deeper':
            result.update(undeeper(value))
        else:
            result[key] = undeeper(value)
    return result


def filter_by_last_level_keys(tree: Dict[str, Any], valid_keys: Map) -> List[Tuple[str, ...]]:
    r"""
    Performs DFS on the Tree and returns values of the nodes on the paths that end with the given keys
    (stored by levels - first level is the root)
    ```
       1
      / \
     2   3
    / \ / \
    4 5 6 7
    ```
    filter_by_last_level_keys(tree, ['4', '7']) => [[1], [2,3]]
    """
    out: List[Tuple[str, ...]] = []

    def recurse(t: Dict[str, Any], vk: Union[Tuple[str, ...], List[str]], acc: List[str]) -> None:
        for key in t.keys():
            if not isinstance(t[key], dict) or t[key] is None:
                if key in vk:
                    nonlocal out
                    out = (
                        [(x,) for x in acc]
                        if len(out) == 0
                        else array_zip(out, [(x,) for x in acc])
                    )
                continue
            else:
                recurse(t[key], vk, acc + [key])

    recurse(tree, valid_keys, [])
    return out


def get_possible_values(
    network: 'BayesianNetwork', possible_values: Dict[str, Union[Tuple[str, ...], List[str]]]
) -> Dict[str, Sequence[str]]:
    """
    Given a `generative-bayesian-network` instance and a set of user constraints, returns an extended
    set of constraints **induced** by the original constraints and network structure
    """

    sets = []
    # For every pre-specified node, compute the "closure" for values of the other nodes
    for key, value in possible_values.items():
        if not isinstance(value, (list, tuple)):
            continue
        if len(value) == 0:
            raise ValueError(
                "The current constraints are too restrictive. No possible values can be found for the given constraints."
            )
        node = network.nodes_by_name[key]
        tree = undeeper(node.node_definition['conditionalProbabilities'])
        zipped_values = filter_by_last_level_keys(tree, value)
        sets.append({**dict(zip(node.parent_names, zipped_values)), key: value})

    # Compute the intersection of all the possible values for each node
    result: Dict[str, Sequence[str]] = {}
    for set_dict in sets:
        for key in set_dict.keys():
            if key in result:
                intersected_values = array_intersection(set_dict[key], result[key])
                if not intersected_values:
                    raise ValueError(
                        "The current constraints are too restrictive. No possible values can be found for the given constraints."
                    )
                result[key] = intersected_values
            else:
                result[key] = set_dict[key]

    return result


def extract_json(path: Path) -> dict:
    """
    Unzips a zip file if the path points to a zip file, otherwise directly loads a JSON file.

    Parameters:
        path: The path to the zip file or JSON file.

    Returns:
        A dictionary representing the JSON content.
    """
    if path.suffix != '.zip':
        # Directly load the JSON file
        with open(path, 'rb') as file:
            return json.loads(file.read())
    # Unzip the file and load the JSON content
    with zipfile.ZipFile(path, 'r') as zf:
        # Find the first JSON file in zip
        try:
            filename = next(file for file in zf.namelist() if file.endswith('.json'))
        except StopIteration:
            return {}  # Broken
        with zf.open(filename) as f:
            # Assuming only one JSON file is needed
            return json.loads(f.read())



from dataclasses import dataclass
from pathlib import Path
from typing import Any, Dict, Iterable, List, Literal, Optional, Tuple, Union

from omar6995_browserforge.bayesian_network import BayesianNetwork, get_possible_values

from .utils import get_browser, get_user_agent, pascalize_headers, tuplify

try:
    import orjson as json
except ImportError:
    import json

try:
    from typing import TypeAlias  # novm
except ImportError:
    from typing_extensions import TypeAlias  # <3.10


"""Constants"""
SUPPORTED_BROWSERS = ('chrome', 'firefox', 'safari', 'edge')
SUPPORTED_OPERATING_SYSTEMS = ('windows', 'macos', 'linux', 'android', 'ios')
SUPPORTED_DEVICES = ('desktop', 'mobile')
SUPPORTED_HTTP_VERSIONS = ('1', '2')
MISSING_VALUE_DATASET_TOKEN = '*MISSING_VALUE*'
HTTP1_SEC_FETCH_ATTRIBUTES = {
    'Sec-Fetch-Mode': 'same-site',
    'Sec-Fetch-Dest': 'navigate',
    'Sec-Fetch-Site': '?1',
    'Sec-Fetch-User': 'document',
}
HTTP2_SEC_FETCH_ATTRIBUTES = {
    'sec-fetch-mode': 'same-site',
    'sec-fetch-dest': 'navigate',
    'sec-fetch-site': '?1',
    'sec-fetch-user': 'document',
}
DATA_DIR: Path = Path(__file__).parent / 'data'
ListOrString: TypeAlias = Union[Tuple[str, ...], List[str], str]


def _parse_version(version: Union[int, float, str, Tuple[int, ...], None]) -> Optional[Tuple[int, ...]]:
    """
    Parse various version formats into a tuple of integers.
    
    Args:
        version: Version in various formats (int, float, str, tuple, or None)
        
    Returns:
        Tuple of integers representing the version, or None if input is None
        
    Examples:
        _parse_version(18) -> (18,)
        _parse_version(18.4) -> (18, 4)
        _parse_version("18.4.1") -> (18, 4, 1)
        _parse_version((18, 4)) -> (18, 4)
    """
    if version is None:
        return None
    
    if isinstance(version, tuple):
        return version
    
    if isinstance(version, int):
        return (version,)
    
    if isinstance(version, float):
        # Handle float like 18.4 -> (18, 4)
        version_str = str(version)
        parts = version_str.split('.')
        return tuple(int(part) for part in parts)
    
    if isinstance(version, str):
        parts = version.split('.')
        return tuple(int(part) for part in parts)
    
    raise ValueError(f"Unsupported version format: {version}")


def _compare_versions(version1: Tuple[int, ...], version2: Tuple[int, ...]) -> int:
    """
    Compare two version tuples.
    
    Args:
        version1: First version tuple
        version2: Second version tuple
        
    Returns:
        -1 if version1 < version2
         0 if version1 == version2
         1 if version1 > version2
    """
    # Pad shorter version with zeros for comparison
    max_len = max(len(version1), len(version2))
    v1_padded = version1 + (0,) * (max_len - len(version1))
    v2_padded = version2 + (0,) * (max_len - len(version2))
    
    if v1_padded < v2_padded:
        return -1
    elif v1_padded > v2_padded:
        return 1
    else:
        return 0


@dataclass
class Browser:
    """Represents a browser specification with name, min/max version, and HTTP version"""

    name: str
    min_version: Optional[Union[int, float, str, Tuple[int, ...]]] = None
    max_version: Optional[Union[int, float, str, Tuple[int, ...]]] = None
    http_version: Union[str, int] = '2'

    def __post_init__(self):
        # Convert http_version to string
        if isinstance(self.http_version, int):
            self.http_version = str(self.http_version)
        
        # Parse versions
        self._min_version_tuple = _parse_version(self.min_version)
        self._max_version_tuple = _parse_version(self.max_version)
        
        # Confirm min_version < max_version
        if (
            self._min_version_tuple is not None
            and self._max_version_tuple is not None
            and _compare_versions(self._min_version_tuple, self._max_version_tuple) > 0
        ):
            raise ValueError(
                f'Browser min version constraint ({self.min_version}) cannot exceed max version ({self.max_version})'
            )

    def version_matches(self, version: Tuple[int, ...]) -> bool:
        """
        Check if a given version matches this browser's version constraints.
        
        Args:
            version: Version tuple to check
            
        Returns:
            True if version matches constraints, False otherwise
        """
        if self._min_version_tuple is not None:
            if _compare_versions(version, self._min_version_tuple) < 0:
                return False
        
        if self._max_version_tuple is not None:
            if _compare_versions(version, self._max_version_tuple) > 0:
                return False
        
        return True


@dataclass
class HttpBrowserObject:
    """Represents an HTTP browser object with name, version, complete string, and HTTP version"""

    name: Optional[str]
    version: Tuple[int, ...]
    complete_string: str
    http_version: str

    @property
    def is_http2(self):
        return self.http_version == '2'


class HeaderGenerator:
    """Generates HTTP headers based on a set of constraints"""

    relaxation_order: Tuple[str, ...] = ('locales', 'devices', 'operatingSystems', 'browsers')

    # Initialize networks
    input_generator_network = BayesianNetwork(DATA_DIR / "input-network.zip")
    header_generator_network = BayesianNetwork(DATA_DIR / "header-network.zip")

    def __init__(
        self,
        browser: Union[ListOrString, Iterable[Browser]] = SUPPORTED_BROWSERS,
        os: ListOrString = SUPPORTED_OPERATING_SYSTEMS,
        device: ListOrString = SUPPORTED_DEVICES,
        locale: ListOrString = 'en-US',
        http_version: Literal[1, 2] = 2,
        strict: bool = False,
    ):
        """
        Initializes the HeaderGenerator with the given options.

        Parameters:
            browser (Union[ListOrString, Iterable[Browser]], optional): Browser(s) or Browser object(s).
            os (ListOrString, optional): Operating system(s) to generate headers for.
            device (ListOrString, optional): Device(s) to generate the headers for.
            locale (ListOrString, optional): List of at most 10 languages for the Accept-Language header. Default is 'en-US'.
            http_version (Literal[1, 2], optional): Http version to be used to generate headers. Defaults to 2.
            strict (bool, optional): Throws an error if it cannot generate headers based on the input. Defaults to False.
        """
        http_ver: str = str(http_version)

        self.options = {
            'browsers': self._prepare_browsers_config(tuplify(browser), http_ver),
            'os': tuplify(os),
            'devices': tuplify(device),
            'locales': tuplify(locale),
            'http_version': http_ver,
            'strict': strict,
        }
        # Loader orders
        self.unique_browsers = self._load_unique_browsers()
        self.headers_order = self._load_headers_order()

    def generate(
        self,
        *,
        browser: Optional[Iterable[Union[str, Browser]]] = None,
        os: Optional[ListOrString] = None,
        device: Optional[ListOrString] = None,
        locale: Optional[ListOrString] = None,
        http_version: Optional[Literal[1, 2]] = None,
        user_agent: Optional[ListOrString] = None,
        strict: Optional[bool] = None,
        request_dependent_headers: Optional[Dict[str, str]] = None,
    ):
        """
        Generates headers using the default options and their possible overrides.

        Parameters:
            browser (Optional[Iterable[Union[str, Browser]]], optional): Browser(s) to generate the headers for.
            os (Optional[ListOrString], optional): Operating system(s) to generate the headers for.
            device (Optional[ListOrString], optional): Device(s) to generate the headers for.
            locale (Optional[ListOrString], optional): Language(s) to include in the Accept-Language header.
            http_version (Optional[Literal[1, 2]], optional): HTTP version to be used to generate headers.
            user_agent (Optional[ListOrString], optional): User-Agent(s) to use.
            request_dependent_headers (Optional[Dict[str, str]], optional): Known values of request-dependent headers.
            strict (Optional[bool], optional): If true, throws an error if it cannot generate headers based on the input.
        """

        options = {
            'browsers': tuplify(browser),
            'os': tuplify(os),
            'devices': tuplify(device),
            'locales': tuplify(locale),
            'http_version': str(http_version) if http_version else None,
            'strict': strict,
            'user_agent': tuplify(user_agent),
            'request_dependent_headers': request_dependent_headers,
        }
        generated: Dict[str, str] = self._get_headers(
            **{k: v for k, v in options.items() if v is not None}
        )
        if (options['http_version'] or self.options['http_version']) == '2':
            return pascalize_headers(generated)
        return generated

    def _get_headers(
        self,
        request_dependent_headers: Optional[Dict[str, str]] = None,
        user_agent: Optional[Iterable[str]] = None,
        **options: Any,
    ) -> Dict[str, str]:
        """
        Generates HTTP headers based on the given constraints.

        Parameters:
            request_dependent_headers (Dict[str, str], optional): Dictionary of request-dependent headers.
            user_agent (Iterable[str], optional): User-Agent value(s).
            **options (Any): Additional options for header generation.

        Returns:
            Dict[str, str]: Dictionary of generated HTTP headers.
        """
        if request_dependent_headers is None:
            request_dependent_headers = {}

        # Process new options
        if 'browsers' in options or (
            # if a unique http_version was passed
            'http_version' in options
            and options['http_version'] != self.options['http_version']
        ):
            self._update_http_version(options)

        header_options = {**self.options, **options}
        possible_attribute_values = self._get_possible_attribute_values(header_options)

        if user_agent:
            # evaluate iterable
            if not isinstance(user_agent, (tuple, list)):
                user_agent = tuple(user_agent)
            http1_values, http2_values = (
                get_possible_values(self.header_generator_network, {'User-Agent': user_agent}),
                get_possible_values(self.header_generator_network, {'user-agent': user_agent}),
            )
        else:
            http1_values, http2_values = {}, {}

        constraints = self._prepare_constraints(
            possible_attribute_values, http1_values, http2_values
        )

        input_sample = self.input_generator_network.generate_consistent_sample_when_possible(
            constraints
        )
        if not input_sample:
            if header_options['http_version'] == '1':
                headers2 = self._get_headers(
                    request_dependent_headers, user_agent, **options, http_version='2'
                )
                return self.order_headers(pascalize_headers(headers2))

            relaxation_index = next(
                (i for i, key in enumerate(self.relaxation_order) if key in options), -1
            )
            if header_options['strict'] or relaxation_index == -1:
                raise ValueError(
                    'No headers based on this input can be generated. Please relax or change some of the requirements you specified.'
                )

            relaxed_options = {**options}
            del relaxed_options[self.relaxation_order[relaxation_index]]
            return self._get_headers(request_dependent_headers, user_agent, **relaxed_options)

        generated_sample = self.header_generator_network.generate_sample(input_sample)
        generated_http_and_browser = self._prepare_http_browser_object(
            generated_sample['*BROWSER_HTTP']
        )

        # Add Accept-Language header
        accept_language_field_name = (
            'accept-language' if generated_http_and_browser.is_http2 else 'Accept-Language'
        )
        generated_sample[accept_language_field_name] = self._get_accept_language_header(
            header_options['locales']
        )

        # Add Sec headers
        if self._should_add_sec_fetch(generated_http_and_browser):
            if generated_http_and_browser.is_http2:
                generated_sample.update(HTTP2_SEC_FETCH_ATTRIBUTES)
            else:
                generated_sample.update(HTTP1_SEC_FETCH_ATTRIBUTES)

        # Ommit connection, close, and missing value headers
        generated_sample = {
            k: v
            for k, v in generated_sample.items()
            if not (
                k.lower() == 'connection'
                and v == 'close'
                or k.startswith('*')
                or v == MISSING_VALUE_DATASET_TOKEN
            )
        }

        # Reorder headers
        return self.order_headers({**generated_sample, **request_dependent_headers})

    def _update_http_version(
        self,
        options: Dict[str, Any],
    ):
        """
        Prepares options when a `browsers` or `http_version` kwarg is passed to .generate.

        Parameters:
            options (Dict[str, Any]): Other arguments.
        """
        if 'http_version' in options:
            http_version = options['http_version']
        else:
            http_version = self.options['http_version']

        if 'browsers' in options:
            options['browsers'] = self._prepare_browsers_config(options['browsers'], http_version)
        else:
            # Create a copy of the class browsers with an updated http_version
            options['browsers'] = [
                (
                    Browser(
                        name=brwsr.name,
                        min_version=brwsr.min_version,
                        max_version=brwsr.max_version,
                        http_version=http_version,
                    )
                    if isinstance(brwsr, Browser)
                    else Browser(name=brwsr, http_version=http_version)
                )
                for brwsr in self.options['browsers']
            ]

    def _prepare_browsers_config(
        self, browsers: Iterable[Union[str, Browser]], http_version: str
    ) -> List[Browser]:
        """
        Prepares the browser configuration based on the given browsers and HTTP version.

        Parameters:
            browsers (Iterable[Union[str, Browser]]): Supported browsers or Browser objects.
            http_version (str): HTTP version ('1' or '2').

        Returns:
            List[Browser]: List of Browser objects.
        """
        return [
            (
                Browser(name=browser, http_version=http_version)
                if isinstance(browser, str)
                else browser
            )
            for browser in browsers
        ]

    def _get_browser_http_options(self, browsers: Iterable[Browser]) -> List[str]:
        """
        Retrieves the browser HTTP options based on the given browser specifications.

        Parameters:
            browsers (Iterable[Browser]): Iterable of Browser objects.

        Returns:
            List[str]: List of browser HTTP options.
        """
        return [
            browser_option.complete_string
            for browser in browsers
            for browser_option in self.unique_browsers
            if browser.name == browser_option.name
            and browser.version_matches(browser_option.version)
            and (not browser.http_version or browser.http_version == browser_option.http_version)
        ]

    def order_headers(self, headers: Dict[str, str]) -> Dict[str, str]:
        """
        Orders the headers based on the browser-specific header order.

        Parameters:
            headers (Dict[str, str]): Dictionary of headers.

        Returns:
            Dict[str, str]: Ordered dictionary of headers.
        """
        # get the browser name
        user_agent = get_user_agent(headers)
        if user_agent is None:
            raise ValueError("Failed to find User-Agent in generated response")
        browser_name = get_browser(user_agent)
        if browser_name is None:
            raise ValueError("Failed to find browser in User-Agent")

        header_order = self.headers_order.get(browser_name)
        # Order headers according to the specific browser's header order
        return (
            {key: headers[key] for key in header_order if key in headers}
            if header_order
            else headers
        )

    def _get_possible_attribute_values(
        self, header_options: Dict[str, Any]
    ) -> Dict[str, List[str]]:
        """
        Retrieves the possible attribute values based on the given header options.

        Parameters:
            header_options (Dict[str, Any]): Dictionary of header options.

        Returns:
            Dict[str, List[str]]: Dictionary of possible attribute values.
        """
        browsers = self._prepare_browsers_config(
            header_options.get('browsers', ()),
            header_options.get('http_version', '2'),
        )
        browser_http_options = self._get_browser_http_options(browsers)

        possible_attribute_values = {
            '*BROWSER_HTTP': browser_http_options,
            '*OPERATING_SYSTEM': header_options.get('os', SUPPORTED_OPERATING_SYSTEMS),
        }
        if 'devices' in header_options:
            possible_attribute_values['*DEVICE'] = header_options['devices']

        return possible_attribute_values

    def _should_add_sec_fetch(self, browser: HttpBrowserObject) -> bool:
        """
        Determines whether Sec-Fetch headers should be added based on the user agent.

        Parameters:
            browser (HttpBrowserObject): Browser object.

        Returns:
            bool: True if Sec-Fetch headers should be added, False otherwise.
        """
        if browser.name == 'chrome' and browser.version[0] >= 76:
            return True
        if browser.name == 'firefox' and browser.version[0] >= 90:
            return True
        if browser.name == 'edge' and browser.version[0] >= 79:
            return True
        return False

    def _get_accept_language_header(self, locales: ListOrString) -> str:
        """
        Generates the Accept-Language header based on the given locales.

        Parameters:
            locales (ListOrString): Locale(s).

        Returns:
            str: Accept-Language header string.
        """
        return ', '.join(
            f"{locale};q={1.0 - index * 0.1:.1f}" for index, locale in enumerate(locales)
        )

    def _load_headers_order(self) -> Dict[str, List[str]]:
        """
        Loads the headers order from the headers-order.json file.

        Returns:
            Dict[str, List[str]]: Dictionary of headers order for each browser.
        """
        headers_order_path = DATA_DIR / "headers-order.json"
        return json.loads(headers_order_path.read_bytes())

    def _load_unique_browsers(self) -> List[HttpBrowserObject]:
        """
        Loads the unique browsers from the browser-helper-file.json file.

        Returns:
            List[HttpBrowserObject]: List of HttpBrowserObject instances.
        """
        browser_helper_path = DATA_DIR / 'browser-helper-file.json'
        unique_browser_strings = json.loads(browser_helper_path.read_bytes())
        return [
            self._prepare_http_browser_object(browser_str)
            for browser_str in unique_browser_strings
            if browser_str != MISSING_VALUE_DATASET_TOKEN
        ]

    def _prepare_constraints(
        self,
        possible_attribute_values: Dict[str, List[str]],
        http1_values: Dict[str, Any],
        http2_values: Dict[str, Any],
    ) -> Dict[str, Iterable[str]]:
        """
        Prepares the constraints for generating consistent samples.

        Parameters:
            possible_attribute_values (Dict[str, List[str]]): Dictionary of possible attribute values.
            http1_values (Dict[str, Any]): Dictionary of HTTP/1 values.
            http2_values (Dict[str, Any]): Dictionary of HTTP/2 values.

        Returns:
            Dict[str, Iterable[str]]: Dictionary of constraints for each attribute.
        """
        return {
            key: tuple(
                filter(
                    lambda x: (
                        self.filter_browser_http(x, http1_values, http2_values)
                        if key == '*BROWSER_HTTP'
                        else self.filter_other_values(x, http1_values, http2_values, key)
                    ),
                    values,
                )
            )
            for key, values in possible_attribute_values.items()
        }

    @staticmethod
    def filter_browser_http(
        value: str, http1_values: Dict[str, Any], http2_values: Dict[str, Any]
    ) -> bool:
        """
        Filters the browser HTTP value based on the HTTP/1 and HTTP/2 values.

        Parameters:
            value (str): Browser HTTP value.
            http1_values (Dict[str, Any]): Dictionary of HTTP/1 values.
            http2_values (Dict[str, Any]): Dictionary of HTTP/2 values.

        Returns:
            bool: True if the value should be included, False otherwise.
        """
        browser_name, http_version = value.split('|')
        return (
            (not http1_values or browser_name in http1_values.get('*BROWSER', ()))
            if http_version == '1'
            else (not http2_values or browser_name in http2_values.get('*BROWSER', ()))
        )

    @staticmethod
    def filter_other_values(
        value: str, http1_values: Dict[str, Any], http2_values: Dict[str, Any], key: str
    ) -> bool:
        """
        Filters the other attribute values based on the HTTP/1 and HTTP/2 values.

        Parameters:
            value (str): Attribute value.
            http1_values (Dict[str, Any]): Dictionary of HTTP/1 values.
            http2_values (Dict[str, Any]): Dictionary of HTTP/2 values.
            key (str): Attribute key.

        Returns:
            bool: True if the value should be included, False otherwise.
        """
        if http1_values or http2_values:
            return value in http1_values.get(key, ()) or value in http2_values.get(key, ())
        return True

    def _prepare_http_browser_object(self, http_browser_string: str) -> HttpBrowserObject:
        """
        Extracts structured information about a browser and HTTP version from a string.

        Parameters:
            http_browser_string (str): HTTP browser string.

        Returns:
            HttpBrowserObject: HttpBrowserObject instance.
        """
        browser_string, http_version = http_browser_string.split('|')
        if browser_string == MISSING_VALUE_DATASET_TOKEN:
            return HttpBrowserObject(
                name=None, version=(), complete_string=MISSING_VALUE_DATASET_TOKEN, http_version=''
            )

        browser_name, version_string = browser_string.split('/')
        version_parts = version_string.split('.')
        version = tuple(int(part) for part in version_parts)
        return HttpBrowserObject(
            name=browser_name,
            version=version,
            complete_string=http_browser_string,
            http_version=http_version,
        )



import shutil
import urllib.request
from concurrent.futures import ThreadPoolExecutor, as_completed
from datetime import datetime, timedelta
from pathlib import Path
from typing import Dict, Iterator

import click

"""
Downloads the required model definitions
"""


ROOT_DIR: Path = Path(__file__).parent

"""Constants for headers and fingerprints data"""
DATA_DIRS: Dict[str, Path] = {
    "headers": ROOT_DIR / "headers/data",
    "fingerprints": ROOT_DIR / "fingerprints/data",
}
DATA_FILES: Dict[str, Dict[str, str]] = {
    "headers": {
        "browser-helper-file.json": "browser-helper-file.json",
        "header-network.zip": "header-network-definition.zip",
        "headers-order.json": "headers-order.json",
        "input-network.zip": "input-network-definition.zip",
    },
    "fingerprints": {
        "fingerprint-network.zip": "fingerprint-network-definition.zip",
    },
}
REMOTE_PATHS: Dict[str, str] = {
    "headers": "https://github.com/apify/fingerprint-suite/raw/master/packages/header-generator/src/data_files",
    "fingerprints": "https://github.com/apify/fingerprint-suite/raw/master/packages/fingerprint-generator/src/data_files",
}


class DownloadException(Exception):
    """Raises when the download fails."""


class DataDownloader:
    """
    Download and extract data files for both headers and fingerprints.
    """

    def __init__(self, **kwargs: bool) -> None:
        self.options = _enabled_flags(kwargs)

    def download_file(self, url: str, path: str) -> None:
        """
        Download a file from the specified URL and save it to the given path.
        """
        # Ensure the directory exists
        Path(path).parent.mkdir(parents=True, exist_ok=True)
        
        with urllib.request.urlopen(url) as resp:  # nosec
            if resp.status != 200:
                raise DownloadException(f"Download failed with status code: {resp.status}")
            with open(path, "wb") as f:
                shutil.copyfileobj(resp, f)

    def download(self) -> None:
        """
        Download and extract data files for both headers and fingerprints.
        """
        futures = {}
        with ThreadPoolExecutor(10) as executor:
            for data_type in self.options:
                for local_name, remote_name in DATA_FILES[data_type].items():
                    url = f"{REMOTE_PATHS[data_type]}/{remote_name}"
                    path = str(DATA_DIRS[data_type] / local_name)
                    future = executor.submit(self.download_file, url, path)
                    futures[future] = local_name
            for f in as_completed(futures):
                try:
                    future.result()
                    click.secho(f"{futures[f]:<30}OK!", fg="green")
                except Exception as e:
                    click.secho(f"Error downloading {local_name}: {e}", fg="red")


def _enabled_flags(flags: Dict[str, bool]) -> Iterator[str]:
    """
    Returns a list of enabled flags based on a given dictionary
    """
    for flag, enabled in flags.items():
        if enabled:
            yield flag


def _get_all_paths(**flags: bool) -> Iterator[Path]:
    """
    Yields all the paths to the downloaded data files
    """
    for data_type in _enabled_flags(flags):
        data_path = DATA_DIRS[data_type]
        for local_name, _ in DATA_FILES[data_type].items():
            yield data_path / local_name


"""
Public download functions
"""


def Download(headers=False, fingerprints=False) -> None:
    """
    Download the required data files
    """
    # Announce that files are being downloaded
    click.secho('Downloading model definition files...', fg='bright_yellow')
    try:
        DataDownloader(headers=headers, fingerprints=fingerprints).download()
    except KeyboardInterrupt:
        print("Download interrupted.")
        Remove()
        exit()


def DownloadIfNotExists(**flags: bool) -> None:
    """
    Download the required data files if they don't exist
    """
    if not IsDownloaded(**flags):
        Download(**flags)


def IsDownloaded(**flags: bool) -> bool:
    """
    Check if the required data files are already downloaded and not older than a month.
    Returns True if all the requested data files are present and not older than a month, False otherwise.
    """
    for path in _get_all_paths(**flags):
        if not path.exists():
            return False

    # Check if the file is older than a month
    file_creation_time = datetime.fromtimestamp(path.stat().st_ctime)
    one_month_ago = datetime.now() - timedelta(weeks=5)
    return file_creation_time >= one_month_ago


def Remove() -> None:
    """
    Deletes all downloaded data files
    """
    for path in _get_all_paths(headers=True, fingerprints=True):
        path.unlink(missing_ok=True)
